package se.gawell.kvtstore.core.inmemory;

import java.rmi.server.UID;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.springframework.beans.factory.annotation.Autowired;

import se.gawell.kvtstore.core.AlphabeticallySortableTimeFormatter;
import se.gawell.kvtstore.core.EntityNotFoundException;
import se.gawell.kvtstore.core.KeyTimeValueStore;
import se.gawell.kvtstore.core.SecondaryIndexValue;
import se.gawell.kvtstore.core.TimeFormatter;
import se.gawell.kvtstore.core.TimeProvider;

@SuppressWarnings("rawtypes")
public class InMemoryKeyTimeValueStore implements KeyTimeValueStore<String> {

	private static final String TIME_INDEX = "time_index";
	private ConcurrentHashMap<Class, ConcurrentHashMap<String, Object>> primaryKeyStore = new ConcurrentHashMap<Class, ConcurrentHashMap<String,Object>>();
	private ConcurrentHashMap<Class, ConcurrentHashMap<String, ConcurrentHashMap<String, Set<Object>>>> indexStore = new ConcurrentHashMap<Class, ConcurrentHashMap<String,ConcurrentHashMap<String,Set<Object>>>>();
	
	private final TimeProvider timeProvider;
	private final TimeFormatter timeFormatter;

	@Autowired
	public InMemoryKeyTimeValueStore(TimeProvider timeProvider) {
		super();
		this.timeProvider = timeProvider;
		timeFormatter = new AlphabeticallySortableTimeFormatter();
	}

	@Override
	public String put(Object object, SecondaryIndexValue... indexes) {
		String generatedPrimaryKey = generatePrimaryKey();
		put(generatedPrimaryKey, object, indexes);
		return generatedPrimaryKey;
	}

	@Override
	public void put(String primaryKey, Object object, SecondaryIndexValue... indexes) {
		ConcurrentHashMap<String, Object> classPrimaryKeyStore = getOrCreateClassPrimaryKeyStore(object.getClass());
		classPrimaryKeyStore.put(primaryKey, object);
		addInIndexStore(object, new SecondaryIndexValue(TIME_INDEX, timeFormatter.format(timeProvider.now())));	
		for(SecondaryIndexValue secondaryIndexValue: indexes) {
			addInIndexStore(object, secondaryIndexValue);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <type> type getById(String primaryKey, Class<? extends Object> type) {
		ConcurrentHashMap<String, Object> classPrimaryKeyStore = getOrCreateClassPrimaryKeyStore(type);
		type result = (type) classPrimaryKeyStore.get(primaryKey);
		if (result == null) {
			throw new EntityNotFoundException("");
		}
		return result;
	}

	@Override
	public <type> List<type> getByTimeInterval(Date startOfIntervalInclusive, Date endOfIntervalExclusive,
			Class<? extends Object> type) {
		String formattedStart = timeFormatter.format(startOfIntervalInclusive);
		String formattedEnd = timeFormatter.format(endOfIntervalExclusive);
		return getByIndexInternal(formattedStart, formattedEnd, type, TIME_INDEX);
	}

	@Override
	public <type> void delete(String primaryKey, Class<? extends Object> type) {
		try {
		Object objectToRemove = getById(primaryKey, type);

		ConcurrentHashMap<String,ConcurrentHashMap<String,Set<Object>>> classIndexBucket = getOrCreateClassIndexBucket(type);
		for (ConcurrentHashMap<String, Set<Object>> index: classIndexBucket.values()) {
			for (String indexValue: index.keySet()) {
				Set<Object> storedValues = index.get(indexValue);
				if (storedValues.contains(objectToRemove)) {
					storedValues.remove(objectToRemove);
				}
			}
		}
		} catch (EntityNotFoundException e) {
			// ok to remove objects that do not exist
		}
		getOrCreateClassPrimaryKeyStore(type).remove(primaryKey);
	}

	@Override
	public <type> List<type> getByIndex(SecondaryIndexValue index, Class<? extends Object> type) {
		return getByIndexInternal(index.value, index.value, type, index.name);
	}

	private void addInIndexStore(Object object, SecondaryIndexValue secondaryIndexValue) {
		ConcurrentHashMap<String, Set<Object>> indexBucket = getOrCreateIndexBucket(object.getClass(), secondaryIndexValue.name);
		Set<Object> indexMapping = indexBucket.get(secondaryIndexValue.value);
		if (indexMapping == null) {
			indexMapping = new HashSet<Object>();
			indexBucket.put(secondaryIndexValue.value, indexMapping);
		}
		indexMapping.add(object);
	}

	@SuppressWarnings("unchecked")
	private ConcurrentHashMap<String,Set<Object>> getOrCreateIndexBucket(Class type,
			String indexName) {
		ConcurrentHashMap<String,ConcurrentHashMap<String,Set<Object>>> classIndexBucket = getOrCreateClassIndexBucket(type);
		ConcurrentHashMap<String,Set<Object>> indexBucket = classIndexBucket.get(indexName);
		if (indexBucket == null) {
			indexBucket = new ConcurrentHashMap<String, Set<Object>>();
			classIndexBucket.put(indexName, indexBucket);
		}
		return indexBucket;
	}

	private ConcurrentHashMap<String,ConcurrentHashMap<String,Set<Object>>> getOrCreateClassIndexBucket(Class<? extends Object> type) {
		ConcurrentHashMap<String,ConcurrentHashMap<String,Set<Object>>> classIndexBucket = indexStore.get(type);
		if (classIndexBucket == null) {
			classIndexBucket = new ConcurrentHashMap<String, ConcurrentHashMap<String,Set<Object>>>();
			indexStore.put(type, classIndexBucket);
		}
		return classIndexBucket;
	}

	private String generatePrimaryKey() {
		return new UID().toString();
	}


	@SuppressWarnings("unchecked")
	private <type> List<type> getByIndexInternal(String startOfIndex, String endOfIndex,
			Class<? extends Object> type, String indexName) {
		LinkedList<type> result = new LinkedList<type>();
		ConcurrentHashMap<String,Set<Object>> indexBucket = getOrCreateIndexBucket(type, indexName);
		for (String indexValue: indexBucket.keySet()) {
			if (indexValue.compareTo(startOfIndex) >= 0
					&& indexValue.compareTo(endOfIndex) <= 0) {
				result.addAll((Collection<? extends type>) indexBucket.get(indexValue));
			}	
		}
		return result;
	}

	private ConcurrentHashMap<String, Object> getOrCreateClassPrimaryKeyStore(Class type) {
		ConcurrentHashMap<String, Object> classPrimaryKeyStore = primaryKeyStore.get(type);
		if (classPrimaryKeyStore == null) {
			classPrimaryKeyStore = new ConcurrentHashMap<String, Object>();
			primaryKeyStore.put(type, classPrimaryKeyStore);
		}
		return classPrimaryKeyStore;
	}

}
